#include "StdAfx.h"
#include "MyArchiveHandler.h"
#include "../../UI/Common/OpenArchive.h"
#include "Windows/PropVariant.h"
#include "Windows/PropVariantConversions.h"

const wchar_t *kExceptionErrorMessage = TEXT("Error: ");
const wchar_t *kUserBreak  = TEXT("Break signaled");

const wchar_t *kMemoryExceptionMessage = TEXT("ERROR: Can't allocate required memory!\n");
const wchar_t *kUnknownExceptionMessage = TEXT("Unknown Error");
const wchar_t *kInternalExceptionMessage = TEXT("Internal Error #");

const wchar_t *kUnsupportedArcTypeMessage = TEXT("Unsupported archive type");
const wchar_t *kUserErrorMessage  = TEXT("Incorrect command line"); // NExitCode::kUserError
const wchar_t *kNoFormats = TEXT("7-Zip cannot find the code that works with archives.");



CMyArchiveHandler::CMyArchiveHandler(void)
{
	m_NumFileInfo = 0;
	m_FileInfo = NULL;
	m_EncryptedHeader = false;
	m_LastErrorMsg = NULL;
	m_ProgressCallback = NULL;
	m_UserData = NULL;
}

CMyArchiveHandler::~CMyArchiveHandler(void)
{
	ClearFileInfo();
	ClearFailureInfo();
}

int CMyArchiveHandler::CmdExec(LPCTSTR baseDir, LPCTSTR cmdLine)
{
	int result;

	delete m_LastErrorMsg;
	m_LastErrorMsg = NULL;
	
	try
	{
		result = Execute(baseDir, cmdLine);
	}
  catch(const CNewException &)
  {
    m_LastErrorMsg = _wcsdup( kMemoryExceptionMessage);
    return (NExitCode::kMemoryError);
  }
  catch(const CArchiveCommandLineException &e)
  {
		size_t strSize = wcslen(kExceptionErrorMessage) + e.Length() + 16;
		m_LastErrorMsg = new TCHAR[strSize];
		swprintf(m_LastErrorMsg, strSize, TEXT("%s%S"), kExceptionErrorMessage, e);
    return (NExitCode::kUserError);
  }
  catch(const CSystemException &systemError)
  {
    if (systemError.ErrorCode == E_OUTOFMEMORY)
    {
      m_LastErrorMsg = _wcsdup(kMemoryExceptionMessage);
      return (NExitCode::kMemoryError);
    }
    if (systemError.ErrorCode == E_ABORT)
    {
      m_LastErrorMsg = _wcsdup(kUserBreak);
      return (NExitCode::kUserBreak);
    }
    UString message;
    NError::MyFormatMessage(systemError.ErrorCode, message);
    m_LastErrorMsg = _wcsdup(message);
    return (NExitCode::kFatalError);
  }
  catch(NExitCode::EEnum &exitCode)
  {
		size_t strSize = wcslen(kInternalExceptionMessage) + 16;
		m_LastErrorMsg = new TCHAR[strSize];
		swprintf(m_LastErrorMsg, strSize, TEXT("%s%d"), kInternalExceptionMessage, exitCode);
    return (exitCode);
  }
  catch(const UString &s)
  {
		size_t strSize = wcslen(kInternalExceptionMessage) + s.Length() + 16;
		m_LastErrorMsg = new TCHAR[strSize];
		swprintf(m_LastErrorMsg, strSize, TEXT("%s%s"), kInternalExceptionMessage, s);
    return (NExitCode::kFatalError);
  }
  catch(const AString &s)
  {
		size_t strSize = wcslen(kInternalExceptionMessage) + s.Length() + 16;
		m_LastErrorMsg = new TCHAR[strSize];
		swprintf(m_LastErrorMsg, strSize, TEXT("%s%s"), kInternalExceptionMessage, s);
    return (NExitCode::kFatalError);
  }
  catch(const char *s)
  {
		size_t strSize = wcslen(kInternalExceptionMessage) + strlen(s) + 16;
		m_LastErrorMsg = new TCHAR[strSize];
		swprintf(m_LastErrorMsg, strSize, TEXT("%s%S"), kInternalExceptionMessage, s);
    return (NExitCode::kFatalError);
  }
  catch(int t)
  {
		size_t strSize = wcslen(kInternalExceptionMessage) + 16;
		m_LastErrorMsg = new TCHAR[strSize];
		swprintf(m_LastErrorMsg, strSize, TEXT("%s%d"), kInternalExceptionMessage, t);
    return (NExitCode::kFatalError);
  }
  catch(...)
  {
    m_LastErrorMsg = _wcsdup(kUnknownExceptionMessage);
    return (NExitCode::kFatalError);
  }
  return  result;	
}

void CMyArchiveHandler::SetCallback(SEVENZIP_PROGRESS_PROC *progressCallback, void *userData)
{
	m_ProgressCallback = progressCallback;
	m_UserData = userData;
}

LPTSTR CMyArchiveHandler::GetLastErrorMsg()
{
	return m_LastErrorMsg;
}

int CMyArchiveHandler::Execute(LPCTSTR baseDir, LPCTSTR cmdLine)
{
  #ifdef _WIN32
  SetFileApisToOEM();
  #endif

	ClearFailureInfo();

  UStringVector commandStrings;
  NCommandLineParser::SplitCommandLine(cmdLine, commandStrings);

  if(commandStrings.Size() == 1)
  {
    return 0;
  }

  CArchiveCommandLineOptions options;

  CArchiveCommandLineParser parser;

  parser.Parse1(commandStrings, options);

  if(options.HelpMode)
  {
    return 0;
  }

  #if defined(_WIN32) && defined(_7ZIP_LARGE_PAGES)
  if (options.LargePages)
  {
    SetLargePageSize();
    NSecurity::EnableLockMemoryPrivilege();
  }
  #endif

	options.StdOutMode = false;
	options.StdInMode = false;

  CStdOutStream &stdStream = options.StdOutMode ? g_StdErr : g_StdOut;

  parser.Parse2(options);

  CCodecs *codecs = new CCodecs;
  CMyComPtr<IUnknown> compressCodecsInfo = codecs;
  HRESULT result = codecs->Load();
  if (result != S_OK) throw CSystemException(result);

  bool isExtractGroupCommand = options.Command.IsFromExtractGroup();

  if (codecs->Formats.Size() == 0 && (isExtractGroupCommand || options.Command.CommandType == NCommandType::kList || options.Command.IsFromUpdateGroup()))
  {
    throw kNoFormats;
	}

  CIntVector formatIndices;
  if (!codecs->FindFormatForArchiveType(options.ArcType, formatIndices))
  {
    throw kUnsupportedArcTypeMessage;
	}


  if (options.Command.CommandType == NCommandType::kInfo || options.Command.CommandType == NCommandType::kBenchmark)
  {
		return 0;
  }

  if (isExtractGroupCommand || options.Command.CommandType == NCommandType::kList)
  {
    if(isExtractGroupCommand)
    {
      CExtractCallbackConsole *ecs = new CExtractCallbackConsole;
      CMyComPtr<IFolderArchiveExtractCallback> extractCallback = ecs;

      ecs->OutStream = &stdStream;

      ecs->PasswordIsDefined = options.PasswordEnabled;
      ecs->Password = options.Password;

      ecs->Init();

      COpenCallbackConsole openCallback;
      openCallback.OutStream = &stdStream;

      openCallback.PasswordIsDefined = options.PasswordEnabled;
      openCallback.Password = options.Password;

      CExtractOptions eo;
      eo.StdOutMode = options.StdOutMode;
      eo.PathMode = options.Command.GetPathMode();
      eo.TestMode = options.Command.IsTestMode();
      eo.OverwriteMode = options.OverwriteMode;
      eo.OutputDir = options.OutputDir;
      eo.YesToAll = options.YesToAll;
      eo.CalcCrc = options.CalcCrc;
      #if !defined(_7ZIP_ST) && !defined(_SFX)
      eo.Properties = options.ExtractProperties;
      #endif
      UString errorMessage;
      CDecompressStat stat;
      HRESULT result = DecompressArchives(
          codecs,
          formatIndices,
          options.ArchivePathsSorted,
          options.ArchivePathsFullSorted,
          options.WildcardCensor.Pairs.Front().Head,
          eo, &openCallback, ecs, errorMessage, stat);
      if (!errorMessage.IsEmpty())
      {
        stdStream << endl << "Error: " << errorMessage;
        if (result == S_OK)
          result = E_FAIL;
      }

      stdStream << endl;
      if (ecs->NumArchives > 1)
        stdStream << "Archives: " << ecs->NumArchives << endl;
      if (ecs->NumArchiveErrors != 0 || ecs->NumFileErrors != 0)
      {
        if (ecs->NumArchives > 1)
        {
          stdStream << endl;
          if (ecs->NumArchiveErrors != 0)
            stdStream << "Archive Errors: " << ecs->NumArchiveErrors << endl;
          if (ecs->NumFileErrors != 0)
            stdStream << "Sub items Errors: " << ecs->NumFileErrors << endl;
        }
        if (result != S_OK) throw CSystemException(result);
        return NExitCode::kFatalError;
      }
      if (result != S_OK) throw CSystemException(result);
      if (stat.NumFolders != 0) stdStream << "Folders: " << stat.NumFolders << endl;
      if (stat.NumFiles != 1 || stat.NumFolders != 0) stdStream << "Files: " << stat.NumFiles << endl;
      stdStream << "Size:       " << stat.UnpackSize << endl << "Compressed: " << stat.PackSize << endl;
    }
    else
    {
      UInt64 numErrors = 0;
      HRESULT result = ListArchives(
          codecs,
          formatIndices,
          options.ArchivePathsSorted,
          options.ArchivePathsFullSorted,
          options.WildcardCensor.Pairs.Front().Head,
          options.PasswordEnabled,
          options.Password,
          numErrors);
      if (numErrors > 0)
      {
        g_StdOut << endl << "Errors: " << numErrors;
        return NExitCode::kFatalError;
      }
      if (result != S_OK) throw CSystemException(result);
    }
  }
  else if(options.Command.IsFromUpdateGroup())
  {
    UString workingDir;

    CUpdateOptions &uo = options.UpdateOptions;
    if (uo.SfxMode && uo.SfxModule.IsEmpty()) uo.SfxModule = TEXT("7zCon.sfx");

    COpenCallbackConsole openCallback;
    openCallback.OutStream = &stdStream;

    #ifndef _NO_CRYPTO
    bool passwordIsDefined =
        options.PasswordEnabled && !options.Password.IsEmpty();
    openCallback.PasswordIsDefined = passwordIsDefined;
    openCallback.Password = options.Password;
    #endif

    CUpdateCallbackConsole callback;

    #ifndef _NO_CRYPTO
    callback.PasswordIsDefined = passwordIsDefined;
    callback.AskPassword = options.PasswordEnabled && options.Password.IsEmpty();
    callback.Password = options.Password;
    #endif
    callback.Init(&stdStream);
    callback.m_ProgressCallback = m_ProgressCallback;
    callback.m_UserData = m_UserData;

    CUpdateErrorInfo errorInfo;

    if (!uo.Init(codecs, formatIndices, options.ArchiveName))
    {
      throw kUnsupportedArcTypeMessage;
		}

		if(wcslen(baseDir) > 0)
		{
			int numCensors = options.WildcardCensor.Pairs.Size();
			for(int i=0; i<numCensors; i++)
			{
				if(options.WildcardCensor.Pairs[i].Prefix.IsEmpty())
				{
					options.WildcardCensor.Pairs[i].Prefix = baseDir;
				}
			}
		}

    HRESULT result = UpdateArchive(codecs, options.WildcardCensor, uo, errorInfo, &openCallback, &callback);

    int exitCode = NExitCode::kSuccess;
    if (callback.CantFindFiles.Size() > 0)
    {
      int numErrors = callback.CantFindFiles.Size();
      for (int i = 0; i < numErrors; i++)
      {
				SEVENZIP_FAILURE_INFO failureInfo;
				
				failureInfo.name = _wcsdup(callback.CantFindFiles[i]);
				failureInfo.cause = SEVENZIP_CANT_FIND_FILE;
				failureInfo.failureCode = callback.CantFindCodes[i];
				m_Failures.Add(failureInfo);
      }
      exitCode = NExitCode::kWarning;
    }

    if (result != S_OK)
    {
      UString message;
      if (!errorInfo.Message.IsEmpty())
      {
        message += errorInfo.Message;
        message += L"\n";
      }
      if (!errorInfo.FileName.IsEmpty())
      {
        message += errorInfo.FileName;
        message += L"\n";
      }
      if (!errorInfo.FileName2.IsEmpty())
      {
        message += errorInfo.FileName2;
        message += L"\n";
      }
      if (errorInfo.SystemError != 0)
      {
        message += NError::MyFormatMessageW(errorInfo.SystemError);
        message += L"\n";
      }
      if (!message.IsEmpty()) stdStream << L"\nError:\n" << message;
      throw CSystemException(result);
    }
    int numErrors = callback.FailedFiles.Size();
    if (numErrors > 0)
    {
      for (int i = 0; i < numErrors; i++)
      {
				SEVENZIP_FAILURE_INFO failureInfo;
				
				failureInfo.name = _wcsdup(callback.FailedFiles[i]);
				failureInfo.cause = SEVENZIP_UPDATE_FAILURE;
				failureInfo.failureCode = callback.FailedCodes[i];
				m_Failures.Add(failureInfo);
      }
      exitCode = NExitCode::kWarning;
    }

    return exitCode;
  }

  return 0;
}

/*int CMyArchiveHandler::ReadArchiveContent(LPCTSTR archivePath, LPCTSTR password)
{
	return ReadContent(archivePath, password);
}*/

UInt32 CMyArchiveHandler::GetNumFiles()
{
	return m_NumFileInfo;
}

SEVENZIP_FILE_INFO *CMyArchiveHandler::GetFileInfo(UInt32 index)
{
	if(index < m_NumFileInfo) return &m_FileInfo[index];
	return NULL;
}

int CMyArchiveHandler::GetNumFailures()
{
	return m_Failures.Size();
}

SEVENZIP_FAILURE_INFO *CMyArchiveHandler::GetFailureInfo(int index)
{
	if(index < m_Failures.Size()) return &m_Failures[index];
	return NULL;
}

bool CMyArchiveHandler::HasEncryptedHeader()
{
	return m_EncryptedHeader;
}

bool CMyArchiveHandler::GetBoolValue(IInArchive *archive, UInt32 index, PROPID propID, bool &value)
{
  NCOM::CPropVariant prop;
  if (archive->GetProperty(index, propID, &prop) != S_OK) return false;
  if (prop.vt != VT_BOOL) return false;
  value = (prop.boolVal != VARIANT_FALSE);
  return true;
}

bool CMyArchiveHandler::GetUInt64Value(IInArchive *archive, UInt32 index, PROPID propID, UInt64 &value)
{
  NCOM::CPropVariant prop;
  if (archive->GetProperty(index, propID, &prop) != S_OK) return false;
  if (prop.vt == VT_EMPTY) return false;
  value = ConvertPropVariantToUInt64(prop);
  return true;
}

bool CMyArchiveHandler::GetFileTime(IInArchive *archive, UInt32 index, PROPID propID, FILETIME &value)
{
  NCOM::CPropVariant prop;
  if(archive->GetProperty(index, propID, &prop) != S_OK) return false;
  if(prop.vt != VT_FILETIME) return false;
  if(prop.filetime.dwLowDateTime == 0 && prop.filetime.dwLowDateTime == 0)
  {
		value = prop.filetime;
		return true;
	}
  FILETIME localFileTime;
  if (!FileTimeToLocalFileTime(&prop.filetime, &localFileTime)) return false;
	value = localFileTime;
  return true;
}

HRESULT CMyArchiveHandler::ListArchives(CCodecs *codecs, const CIntVector &formatIndices,
    UStringVector &archivePaths, UStringVector &archivePathsFull,
    const NWildcard::CCensorNode &wildcardCensor,
    bool &passwordEnabled, UString &password,
    UInt64 &numErrors)
{
  numErrors = 0;

  for (int i = 0; i < archivePaths.Size(); i++)
  {
    const UString &archiveName = archivePaths[i];
    NFile::NFind::CFileInfoW fi;
    if (!fi.Find(archiveName) || fi.IsDir())
    {
      numErrors++;
      continue;
    }

    CArchiveLink archiveLink;

    COpenCallbackConsole openCallback;
    openCallback.OutStream = &g_StdOut;

    #ifndef _NO_CRYPTO

    openCallback.PasswordIsDefined = passwordEnabled;
    openCallback.Password = password;

    #endif

    HRESULT result = archiveLink.Open2(codecs, formatIndices, false, NULL, archiveName, &openCallback);
   	m_EncryptedHeader = openCallback.Open_WasPasswordAsked();

    if (result != S_OK)
    {
      if (result == E_ABORT) return result;
      numErrors++;
      continue;
    }

    for (int v = 0; v < archiveLink.VolumePaths.Size(); v++)
    {
      int index = archivePathsFull.FindInSorted(archiveLink.VolumePaths[v]);
      if (index >= 0 && index > i)
      {
        archivePaths.Delete(index);
        archivePathsFull.Delete(index);
      }
    }

    const CArc &arc = archiveLink.Arcs.Back();
    IInArchive *archive = arc.Archive;

    UInt32 numItems;
    RINOK(archive->GetNumberOfItems(&numItems));
    
		ClearFileInfo();
		m_FileInfo = new SEVENZIP_FILE_INFO[numItems];
		if(!m_FileInfo) return E_OUTOFMEMORY;
		m_NumFileInfo = numItems;
		UInt32 fileInfoIndex = 0;
    
    for(UInt32 i = 0; i < numItems; i++)
    {
			UString filePath;
      HRESULT res = arc.GetItemPath(i, filePath);
			if(filePath.Length() == 0) continue;

			bool isDir;
			RINOK(IsArchiveItemFolder(archive, i, isDir));
      if (!wildcardCensor.CheckPath(filePath, !isDir)) continue;

			m_FileInfo[fileInfoIndex].isDir = isDir;

			m_FileInfo[fileInfoIndex].name = new TCHAR[filePath.Length() + 1];
			wcscpy(m_FileInfo[fileInfoIndex].name, filePath);
	    
			if (!GetBoolValue(archive, i, kpidEncrypted, m_FileInfo[fileInfoIndex].isEncrypted)) m_FileInfo[fileInfoIndex].isEncrypted = false;
			if (!GetUInt64Value(archive, i, kpidSize, m_FileInfo[fileInfoIndex].size)) m_FileInfo[fileInfoIndex].size = 0;
			if (!GetUInt64Value(archive, i, kpidPackSize, m_FileInfo[fileInfoIndex].packedSize)) m_FileInfo[fileInfoIndex].packedSize = 0;
			if (!GetFileTime(archive, i, kpidMTime, m_FileInfo[fileInfoIndex].modifiedTime))
			{
				m_FileInfo[fileInfoIndex].modifiedTime.dwLowDateTime = 0;
				m_FileInfo[fileInfoIndex].modifiedTime.dwHighDateTime = 0;
			}
			fileInfoIndex++;
		}
		m_NumFileInfo = fileInfoIndex;
  }

  return S_OK;
}

void CMyArchiveHandler::ClearFileInfo()
{
	if(!m_FileInfo) return;
	
	for(UInt32 i = 0; i < m_NumFileInfo; i++)
	{
		delete[] m_FileInfo[i].name;
	}
	delete[] m_FileInfo;
	m_FileInfo = NULL;
	m_NumFileInfo = 0;
	m_EncryptedHeader = false;
}

void CMyArchiveHandler::ClearFailureInfo()
{
	for(int i = 0; i < m_Failures.Size(); i++)
	{
		free(m_Failures[i].name);
	}
	m_Failures.Clear();
}


